
--
-- Iterator Sequences, based on ideas of
--
-- Wim Couwenberg
-- Steve Donovan
-- Mark Hamburg
--
-- January 2, 2010
--
-- See the file "sequence_tests.lua" for examples
--

module "sequence"

local _iterator = {}
iterator = _iterator

function seq(iter, state, key)
    local apply, self

    function apply(next_iter, ...)
        if next_iter == "results" then
            return ...
        elseif next_iter ~= nil then
            iter = next_iter
        end
        return self, state, key
    end

    function self(s, k, ...)
        if s == state then
            return iter(s, k)
        elseif s == _iterator then
            return iter, state, key
        end
        return apply(s(self, k, ...))
    end

    return self, state, key
end

local function foreach_helper(op, k, ...)
    if k ~= nil then
        op(k, ...)
    end
    return k
end

function foreach(seq, op, ...)
    local iter, state, key = seq(_iterator)
    repeat
        key = foreach_helper(op, iter(state, key))
    until key == nil
    return "results", ...
end

function filter(seq, test)
    local iter = seq(_iterator)

    local function helper(state, key, ...)
        if key ~= nil then
            if test(key, ...) then
                return key, ...
            end
            return helper(state, iter(state, key))
        end
    end

    return function(s, k)
        return helper(s, iter(s, k))
    end
end

function filterv(seq, test)
    local iter = seq(_iterator)

    local function helper(state, key, ...)
        if key ~= nil then
            if test(...) then
                return key, ...
            end
            return helper(state, iter(state, key))
        end
    end

    return function(s, k)
        return helper(s, iter(s, k))
    end
end

function map(seq, op)
    local iter = seq(_iterator)
    local saved_key

    local function helper(key, ...)
        saved_key = key
        if key ~= nil then
            return op(key, ...)
        end
    end

    return function(s, k)
        return helper(iter(s, saved_key or k))
    end
end

function mapv(seq, op)
    local iter = seq(_iterator)

    local function helper(key, ...)
        if key ~= nil then
            return key, op(...)
        end
    end

    return function(s, k)
        return helper(iter(s, k))
    end
end

function take(seq, n)
    local iter = seq(_iterator)

    return function(s, k)
        if n > 0 then
            n = n - 1
            return iter(s, k)
        end
    end
end

function skip(seq, n)
    local iter = seq(_iterator)

    return function(s, k)
        while n > 0 do
            n = n - 1
            k = iter(s, k)
            if k == nil then
                return
            end
        end
        return iter(s, k)
    end
end

local function _empty()
end

function key(seq)
    seq(mapv, _empty)
end

local function _value(key, ...)
    return ...
end

function value(seq)
    seq(map, _value)
end

function append(list)
    return function(value)
        list[#list + 1] = value
    end, list
end

function appendv(list)
    return function(key, value)
        list[#list + 1] = value
    end, list
end

function insert(table)
    return function(key, value)
        table[key] = value
    end, table
end
