
local s = require "sequence"

local numbers = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10}
local words = {"aap", "noot", "mies", "wim", "jet", "vuur"}
local mixed = {1, "aap", 2, "noot", 3, "mies", "4", "wim"}
local tab = {
    aap = 1,
    noot = 2,
    mies = 3,
    wim = 4,
    jet = 5,
    vuur = 6,
}

-- some random text ;-)
local text = [[IMHO the whole thing looks a lot like a solution in search of a problem ...]]

-- seq should work as a normal iterator
print("test 1")
print("------")

for i, v in s.seq(ipairs(numbers)) do
    print(i, v)
end

for k, v in s.seq(pairs(tab)) do
    print(k, v)
end

for w in s.seq(string.gmatch(text, "%w+")) do
    print(w)
end

-- (iterator) should return a true iterator
print("test 2")
print("------")

for i, v in s.seq(ipairs(numbers)) (s.iterator) do
    print(i, v)
end

for k, v in s.seq(pairs(tab)) (s.iterator) do
    print(k, v)
end

for w in s.seq(string.gmatch(text, "%w+")) (s.iterator) do
    print(w)
end

-- foreach must work like iteration
print("test 3")
print("------")

s.seq(ipairs(numbers)) (s.foreach, print)
s.seq(pairs(tab)) (s.foreach, print)
s.seq(string.gmatch(text, "%w+")) (s.foreach, print)

-- filter odd numbers or words of odd length
print("test 4")
print("------")

local function is_odd(i, v)
    return v%2 == 1
end

local function has_odd_length(w)
    return #w%2 == 1
end

s.seq(ipairs(numbers)) (s.filter, is_odd) (s.foreach, print)
s.seq(pairs(tab)) (s.filter, is_odd) (s.foreach, print)
s.seq(string.gmatch(text, "%w+")) (s.filter, has_odd_length) (s.foreach, print)

-- filter values
print("test 5")
print("------")

local function is_odd(v)
    return v%2 == 1
end

s.seq(ipairs(numbers)) (s.filterv, is_odd) (s.foreach, print)
s.seq(pairs(tab)) (s.filterv, is_odd) (s.foreach, print)

-- map should act on (key, value) without disturbing iteration
print("test 6")
print("------")

local function map1(i, v)
    return -i, v^2
end

local function map2(k, v)
    return k .. k, v^2
end

s.seq(ipairs(numbers)) (s.map, map1) (s.foreach, print)
s.seq(pairs(tab)) (s.map, map2) (s.foreach, print)
s.seq(string.gmatch(text, "%w+")) (s.map, string.upper) (s.foreach, print)

-- mapv must operate only on values
print("test 7")
print("------")

local function map3(v)
    return v^2
end

s.seq(ipairs(numbers)) (s.mapv, map3) (s.foreach, print)
s.seq(pairs(tab)) (s.mapv, map3) (s.foreach, print)

-- take must only take the first given number of items
print("test 8")
print("------")

s.seq(ipairs(numbers)) (s.take, 5) (s.foreach, print)
s.seq(pairs(tab)) (s.take, 3) (s.foreach, print)
s.seq(string.gmatch(text, "%w+")) (s.take, 7) (s.foreach, print)

-- skip must skip the given number of items
print("test 9")
print("------")

s.seq(ipairs(numbers)) (s.skip, 5) (s.foreach, print)
s.seq(pairs(tab)) (s.skip, 3) (s.foreach, print)
s.seq(string.gmatch(text, "%w+")) (s.skip, 10) (s.foreach, print)

-- key should strip all values, leaving only the key
print("test 10")
print("-------")

s.seq(ipairs(numbers)) (s.key) (s.foreach, print)
s.seq(pairs(tab)) (s.key) (s.foreach, print)

-- value should strip the (first) key without affecting the iteration
print("test 11")
print("-------")

s.seq(ipairs(numbers)) (s.value) (s.foreach, print)
s.seq(pairs(tab)) (s.value) (s.foreach, print)

-- append should append keys in a list
print("test 12")
print("-------")

local list
list = s.seq(ipairs(numbers)) (s.value) (s.foreach, s.append {})
print(unpack(list))
list = s.seq(pairs(tab)) (s.value) (s.foreach, s.append(list))
print(unpack(list))
list = s.seq(string.gmatch(text, "%w+")) (s.foreach, s.append(list))
print(unpack(list))

-- appendv should append values in a list
print("test 13")
print("-------")

list = nil
list = s.seq(ipairs(numbers)) (s.foreach, s.appendv {})
print(unpack(list))
list = s.seq(pairs(tab)) (s.foreach, s.appendv(list))
print(unpack(list))

-- insert should insert key value pairs in a table
print("test 14")
print("-------")

local copy = s.seq(ipairs(numbers)) (s.foreach, s.insert {})
s.seq(pairs(copy)) (s.foreach, print)
copy = s.seq(pairs(tab)) (s.foreach, s.insert(copy))
s.seq(pairs(copy)) (s.foreach, print)

-- miscellaneous nonsense...
print("test 15")
print("-------")

local function isnumber(x)
    return type(x) == "number"
end

s.seq(ipairs(mixed)) (s.mapv, tonumber) (s.filterv, isnumber) (s.foreach, print)

local function to_number(seq)
    seq(s.mapv, tonumber) (s.filterv, isnumber)
end

s.seq(ipairs(mixed)) (to_number) (s.foreach, print)

local msg = s.seq(string.gmatch(text, "%w+")) (s.skip, 3) (s.take, 5) (s.foreach, s.append {})
print(table.concat(msg, " "))

local function reverse(seq)
    local items = seq(s.foreach, s.append {})
    local busy

    return function(s, k)
        if not busy then
            k = 0
            busy = true
        end
        if k < #items then
            return k + 1, items[#items - k]
        end
    end
end

s.seq(string.gmatch(text, "%w+")) (s.take, 10) (reverse) (s.mapv, function(x) return string.upper(x) .. "!" end) (s.foreach, print)
